package Jcg.mesh;

/**
 * Efficient implementation of an hash-map for storing edges of a triangulation. <p>
 * 
 * Remarks: <p>
 * 1) given an edge (u, v), the hash code is 'u*v'
 * 2) the table stores triples (u, v, eIndex), where (u, v) is the key, and 'eIndex' is the value associated to that key
 * 
 * @param n size of the map
 */
public class HashMapEdges {
    private int[] table; // table of size n*3: it stores, for each entry, 3 values (key1, key2, value)
    private final int n;
    
    int size=0;
    
    /**
     * 
     * @param n size of the map
     */
    public HashMapEdges(int n) {
    	this.n=n;
          table = new int[n*3];
          for (int i = 0; i < n*3; i++)
                table[i] = -1;
    }

    /**
     * The 'key' is assumed to be the unordered pair (u, v)
     */
    public int get(int u, int v) {
    	int hash = (Math.floorMod(u*v, n));

    	int min, max;
    	if(u<=v) {
    		min=u;
    		max=v;
    	}
    	else {
    		min=v;
    		max=u;
    	}
    	
    	//int count=0;
    	while (table[hash*3+2]!=-1 && (table[hash*3]!=min || table[hash*3+1]!=max)) {
    		//hash = Math.floorDiv(hash + 1, n);
    		hash=(hash+1)%n;
    	}
    	
    	if (table[hash*3+2] == -1)
    		return -1;
    	else
    		return table[hash*3+2];

    }

    /**
     * Insert the triple (u, v, eIndex) in the table, if not already existing.
     * 
     * -) If the (u, v) already exists: return the corresponding value
     * -) otherwise, insert (u, v, eIndex) in the table and return -1
     */
    public int put(int u, int v, int eIndex) {
    	//System.out.println("inserting ("+u+", "+v+", "+eIndex+") ");
    	if(size>=n) {
    		System.err.println("Warning: the hash map is full");
    		return -1;
    	}
    	
    	int hash = (Math.floorMod(u*v, n)); // needs Java 8 or higher
    	
    	int min, max;
    	if(u<=v) {
    		min=u;
    		max=v;
    	}
    	else {
    		min=v;
    		max=u;
    	}

    	while (table[hash*3+2]!=-1 && (table[hash*3]!=min || table[hash*3+1]!=max))
    		//hash = Math.floorDiv(hash + 1, n);
    		hash=(hash+1)%n;
    	
    	if (table[hash*3+2] == -1) { // insert the (key, value) in the table
    		table[hash*3]=min;
    		table[hash*3+1]=max;
    		table[hash*3+2]=eIndex;
    		
    		size++;
    		return -1;
    	}
    	else {
    		//System.out.println("\t value already exists");
    		return table[hash*3+2]; // return the value stored with the key (u, v)
    	}

    	/*while (table[hash] != null && table[hash].getKey() != key)

                hash = (hash + 1) % TABLE_SIZE;

          table[hash] = new HashEntry(key, value);*/

    }

    public static void main(String[] args) {
    	int k=10000; // nVertices=10M
    	int nMax=3*k; // 
    	
    	System.out.print("Testing efficient hash map for storing edges...");
    	HashMapEdges map=new HashMapEdges(k*4);
    	
    	/*int[][] testArray=new int[k][k];
    	for(int i=0;i<k;i++) {
    		for(int j=0;j<k;j++)
    			testArray[i][j]=-1;
    	}*/
    	
    	int nValuesMap=0;
    	int nValuesArray=0;
    	for(int i=0;i<nMax;i++) {
    		int u=(int)(Math.random()*k);
    		int v=(int)(Math.random()*k);
    		int eIndex=(int)(Math.random()*k);
    		eIndex=2*u+2*v;
    		
    		//System.out.println("\ninserting ("+u+", "+v+", "+eIndex+") ");
    		int inserted=map.put(u, v, eIndex);
    		if(inserted==-1)
    			nValuesMap++;
    		
    		//map.print();
    		
    		int checkInsertedKey=map.get(u, v);
    		if(checkInsertedKey==-1)
    			System.err.println("Error: key ("+u+", "+v+", "+eIndex+") not stored");
    		else if(checkInsertedKey!=eIndex)
    			System.err.println(i+": Error: the stored value is wrong: "+eIndex+" - "+checkInsertedKey);
    	}
    	
    	System.out.println("ok ["+nValuesMap+" values inserted in hashmap], "+nValuesArray+" values in array");
    }
    
    public void print() {
    	System.out.println("Hash map of size :"+this.n);
    	
    	for(int i=0;i<this.n;i++) {
    		int index=i*3;
    		System.out.println("("+this.table[index]+","+this.table[index+1]+"; "+this.table[index+2]+") ");
    	}
    }

}
