package Jcg.mesh;

import java.io.File;
import java.util.*;

import Jcg.geometry.*;
import Jcg.polyhedron.*;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2014)
 * This class provides methods for building a polyhedral mesh (Polyhedron_3 class)
 */
public class MeshBuilder<X extends Point_> {
		
	/**
	 * Efficient construction of a Polyhedron (Half-edge DS) for a triangle mesh (from a share vertex representation)
	 * Warning: old version, it works only for triangle meshes
	 */
	public Polyhedron_3<X> createMesh(X[] points, int[] faceDegrees, int[][] faces) {
		this.getMemoryUsage();
		System.gc(); // call the garbage collector
		System.gc();
		System.gc();
		System.gc();
		System.out.print("Building a (pointer based) halfedge representation of a surface mesh (from a shared vertex representation)...");
		long startTime=System.nanoTime(), endTime; // for evaluating time performances

		int sizeHalfedges=0; // counts the number of half-edges (twice the sum of face degrees)
        int nInsertedHalfedges=0; // counts the number of inserted half-edges
        
        for(int i=0;i<faceDegrees.length;i++) {
        	sizeHalfedges=sizeHalfedges+faceDegrees[i];
        }

		// allows to efficient retrieving an inserted half-edge
    	HashMap<HalfedgePair, Halfedge<X>> insertedHalfedges=new HashMap(sizeHalfedges);
    	
    	Polyhedron_3<X> result=new Polyhedron_3<X>(points.length,sizeHalfedges,faceDegrees.length);
        
    	int i;
        for(i=0;i<points.length;i++){ // add geometric points
    		Vertex<X> v=new Vertex<X>();
        	v.index=i;
            result.vertices.add(v); // adding a vertex
            
            // store its geometric coordinates
            int index=i*3;
        	result.coordF[index]=(float)points[i].getCartesian(0).doubleValue();
        	result.coordF[index+1]=(float)points[i].getCartesian(1).doubleValue();
        	result.coordF[index+2]=(float)points[i].getCartesian(2).doubleValue();
        }
        
        for(i=0;i<faceDegrees.length;i++) {
        	int d=3; // degree of face i
            
            // creating and storing the 'd' half-edges in a face
            Halfedge[] _edges=new Halfedge[d];
            for(int j=0;j<d;j++) {
            	Halfedge<X> e=new Halfedge<X>();
            	result.halfedges.add(e);
            	_edges[j]=e; // storing the edges of the face
            	
            	// setting incident (target) vertex
            	e.setVertex(result.vertices.get(faces[i][(j+1)%d]));
            	// half-edges must be incident to the current face
            }
            
            // setting incidence relations between half-edges in the same face
            for(int j=0;j<d;j++) {
            	int indNext=(j+1)%d; // index of next half-edge
            	int indPrev; // index of previous half-edge
            	if(j>0) indPrev=j-1;
            	else indPrev=d-1;
            	
            	_edges[j].setNext(_edges[indNext]);
            }
            
            // setting opposite half-edge
            for(int j=0;j<d;j++) {
            	HalfedgePair pair0=new HalfedgePair(faces[i][j], faces[i][(j+1)%d]);
            	if(insertedHalfedges.containsKey(pair0)==false) {
            		insertedHalfedges.put(pair0, _edges[j]);
            		//System.out.println("inserted edge ");
            		nInsertedHalfedges++;
            	}
            	else {
            		//System.out.println("edge already inserted");
            		Halfedge<X> eOpposite=insertedHalfedges.get(pair0);
            		_edges[j].setOpposite(eOpposite);
            		eOpposite.setOpposite(_edges[j]);
            	}
            }
            
        }
        
        // setting the vertex incident to a half-edge
		for(Halfedge pedge: result.halfedges) {
        	Vertex<X> pvertex=pedge.vertex;
        	pvertex.setEdge(pedge);
        }
		
		endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");
		
    	this.getMemoryUsage();
        return result;        
    }
	
	/**
	 * Efficient construction of a polyhedral mesh (Half-edge DS) 
	 * (from a share vertex representation)
	 */
	Polyhedron_3<X> createMeshOld(X[] points, int[] faceDegrees, int[][] faces) {   
		System.out.print("Building a (pointer based) halfedge representation of a triangle mesh...");
		
		int sizeHalfedges=0; // count the number of half-edges (twice the sum of face degrees)s
        int nInsertedHalfedges=0; // counts the number of inserted halfedges

        // allows to efficient retrieving an inserted halfedge
    	HashMap<HalfedgePair, Halfedge<X>> insertedHalfedges=new HashMap(sizeHalfedges);
    	
    	Polyhedron_3<X> result=new Polyhedron_3<X>(points.length,sizeHalfedges,faceDegrees.length);
        
    	int i;
    	for(i=0;i<points.length;i++){ // add geometric points
    		Vertex<X> v=new Vertex<X>();
        	v.index=i;
            result.vertices.add(v); // adding a vertex
            
            // store its geometric coordinates
            int index=i*3;
        	result.coordF[index]=(float)points[i].getCartesian(0).doubleValue();
        	result.coordF[index+1]=(float)points[i].getCartesian(1).doubleValue();
        	result.coordF[index+2]=(float)points[i].getCartesian(2).doubleValue();
        }
    	
    	for(i=0;i<faceDegrees.length;i++) {         
            
            // creating and storing the three halfedges in a face
            Halfedge<X> e0=new Halfedge<X>();
            Halfedge<X> e1=new Halfedge<X>();
            Halfedge<X> e2=new Halfedge<X>();
            result.halfedges.add(e0);
            result.halfedges.add(e1);
            result.halfedges.add(e2);
            
            // setting incident (target vertex)
            e0.setVertex(result.vertices.get(faces[i][1]));
            e1.setVertex(result.vertices.get(faces[i][2]));
            e2.setVertex(result.vertices.get(faces[i][0]));
            
            // setting first opposite halfedge
            HalfedgePair pair0=new HalfedgePair(faces[i][0], faces[i][1]);
            if(insertedHalfedges.containsKey(pair0)==false) {
            	insertedHalfedges.put(pair0, e0);
            	//System.out.println("inserted edge ");
            	nInsertedHalfedges++;
            }
            else {
            	//System.out.println("edge already inserted");
            	Halfedge<X> eOpposite=insertedHalfedges.get(pair0);
            	e0.setOpposite(eOpposite);
            	eOpposite.setOpposite(e0);
            }
            // setting second opposite halfedge
            HalfedgePair pair1=new HalfedgePair(faces[i][1], faces[i][2]);
            if(insertedHalfedges.containsKey(pair1)==false) {
            	insertedHalfedges.put(pair1, e1);
            	nInsertedHalfedges++;
            }
            else {
            	//System.out.println("edge already inserted");
            	Halfedge<X> eOpposite=insertedHalfedges.get(pair1);
            	e1.setOpposite(eOpposite);
            	eOpposite.setOpposite(e1);
            }
            // setting third opposite halfedge
            HalfedgePair pair2=new HalfedgePair(faces[i][2], faces[i][0]);
            if(insertedHalfedges.containsKey(pair2)==false) {
            	insertedHalfedges.put(pair2, e2);
            	nInsertedHalfedges++;
            }
            else {
            	//System.out.println("edge already inserted");
            	Halfedge<X> eOpposite=insertedHalfedges.get(pair2);
            	e2.setOpposite(eOpposite);
            	eOpposite.setOpposite(e2);
            }
            
        }
        
        // setting the vertex incident to a halfedge
		for(i=0;i<result.halfedges.size();i++) {
        	Halfedge<X> pedge=(Halfedge<X>)result.halfedges.get(i);
        	Vertex<X> pvertex=pedge.vertex;
        	pvertex.setEdge(pedge);
        }
		System.out.println("done");
		
        return result;        
    }
	
	/**
	 * Print the edge (the index of its extremities)
	 * Useful for debugging
	 */
	private String printEdge(Halfedge<X> e) {
		if(e==null) return "null";
		return "("+e.getPrev().getVertex().index+" - "+e.getVertex().index+")";
    }
	
	/**
	 * Convert an array of 3D points to an array of 2D points (by orthogonal projection on the z=0 hyperplane)
	 */
    public static Point_2[] Point3DToPoint2D(Point_3[] points) {
    	Point_2[] result=new Point_2[points.length];
    	for(int i=0;i<points.length;i++) {
    		result[i]=new Point_2(points[i].getX().doubleValue(), points[i].getY().doubleValue());
    	}
    	return result;
    }
    
    public static void getMemoryUsage() {
    	/* Total number of processors or cores available to the JVM */
    	  System.out.println("\n\t---Available processors (cores): " + 
    	  Runtime.getRuntime().availableProcessors());

    	  /* Total amount of free memory available to the JVM */
    	  System.out.println("Free memory (bytes): " + 
    	  Runtime.getRuntime().freeMemory());

    	  /* This will return Long.MAX_VALUE if there is no preset limit */
    	  long maxMemory = Runtime.getRuntime().maxMemory();
    	  /* Maximum amount of memory the JVM will attempt to use */
    	  System.out.println("Maximum memory (bytes): " + 
    	  (maxMemory == Long.MAX_VALUE ? "no limit" : maxMemory));

    	  /* Total memory currently in use by the JVM */
    	  System.out.println("Total memory (bytes): " + 
    	  Runtime.getRuntime().totalMemory());

    	  /* Get a list of all filesystem roots on this system */
    	  File[] roots = File.listRoots();

    	  /* For each filesystem root, print some info */
    	  for (File root : roots) {
    	    System.out.println("File system root: " + root.getAbsolutePath());
    	    System.out.println("Total space (bytes): " + root.getTotalSpace());
    	    System.out.println("Free space (bytes): " + root.getFreeSpace());
    	    System.out.println("Usable space (bytes): " + root.getUsableSpace());
    	  }

    	  System.out.println("\t----------------\n");
    }
            
}
