package Jcg.mesh;

import java.awt.Color;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.HashMap;

import Jcg.geometry.*;
import Jcg.polyhedron.*;

/**
 * A vertex shared representation of a mesh.
 * For each face we store the indices of the incident vertices
 *
 * @author Luca Castelli Aleardi (INF555, 2012)
 *
 */
public class SharedVertexRepresentation {
	    
	public int sizeVertices;
	public int sizeFaces;
	public int sizeHalfedges;
	
    public int[][] faces;
    public int[] faceDegrees;
    public Point_3[] points;

    public Color[] faceColors=null;
    public Color[] edgesColors=null;
    
    /**
     * Create a shared vertex representation
     */
    public SharedVertexRepresentation(int n, int f, int he, int[][] faces, int[] faceDegrees, Point_3[] points) {
    	System.out.print("Creating a shared vertex representation...");
    	
    	this.sizeVertices=n;
    	this.sizeFaces=f;
    	this.sizeHalfedges=he;
    	this.faces=faces;
    	this.faceDegrees=faceDegrees;
    	this.points=points;
    	
    	System.out.println("done");
    }

    /**
     * Create a shared vertex representation from a collection of 3D triangles
     */
    public SharedVertexRepresentation(Collection<Triangle_3> faces) {
    	System.out.print("Creating a shared vertex representation (from collection of triangles)...");
    	this.faces = new int[faces.size()][];
    	faceDegrees = new int[faces.size()];
    	faceColors = new Color[faces.size()];
    	LinkedHashMap<Point_3,Integer> vert = new LinkedHashMap<Point_3,Integer> ();

    	int ind=0;
    	int indF = 0;
    	for (Triangle_3 f : faces) {
    		for (int i=0; i<3; i++)
    			if (!vert.containsKey(f.vertex(i)))
    				vert.put(f.vertex(i), ind++);
    		faceDegrees[indF] = 3;
    		faceColors[indF] = Color.gray;
    		this.faces[indF++] = new int[]{vert.get(f.vertex(0)), vert.get(f.vertex(1)), vert.get(f.vertex(2))};
    	}    	
    	points = vert.keySet().toArray(new Point_3[0]);
    	System.out.println("done");
    }
    
    /**
     * Create a mesh representation from an off file (using TC library)
     */
    public SharedVertexRepresentation(String filename){
    	System.out.print("Reading a mesh from file in OFF format (shared vertex representation): ");
    	System.out.println(filename);
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	double x, y, z;
    	Jcg.io.IO.readTextFile(filename);
    	String line;
            
            line=Jcg.io.IO.readLine(); // first line is empty
            line=Jcg.io.IO.readLine();
            String[] w=Jcg.io.IO.wordsFromString(line);
            sizeVertices=Integer.parseInt(w[0]);
            sizeFaces=Integer.parseInt(w[1]);            
            
            points=new Point_3[sizeVertices];
            faceDegrees=new int[sizeFaces];
            faces=new int[sizeFaces][];
            
            int i=0;
            Point_3 point;
            System.out.print("\tReading vertices...");
            
            while(i<sizeVertices) {
                line=Jcg.io.IO.readLine();
                w = Jcg.io.IO.wordsFromString(line);
                if(w!=null && w.length>0 && w[0].charAt(0)!='#') { // skip empty lines an comments
                	x=(new Double(w[0])).doubleValue();
                	y=(new Double(w[1])).doubleValue();
                	z=(new Double(w[2])).doubleValue();
                
                	point=new Point_3(x,y,z);
                	points[i]=point;
                	i++;
                }
            }
            System.out.println("done "+sizeVertices+" vertices");
            
            System.out.print("\tReading face degrees...");
            i=0;
            while(i<sizeFaces){
            	line = Jcg.io.IO.readLine();
            	if(line == null)  {
            		throw new Error("error: end of file reached before reading all faces");
            	}
            	String[] words=Jcg.io.IO.wordsFromString(line);
            	if(words!=null && words.length>1) {            		
            		faceDegrees[i]=Integer.parseInt(words[0]); // first element encodes the degree
            		faces[i]=new int[faceDegrees[i]];
            		
            		for(int j=0; j<faceDegrees[i]; j++) {
            			faces[i][j]=Integer.parseInt(words[j+1]);
            			sizeHalfedges++;
                    }
            		i++;
                }         
            }
            System.out.println("done "+sizeFaces+" faces");
            Jcg.io.IO.readStandardInput();
            
            endTime=System.nanoTime();
            double duration=(double)(endTime-startTime)/1000000000.;
            System.out.print("Mesh loaded from OFF format");
            System.out.println(" ("+duration+" seconds)");
    }
	
    /**
     * Create a shared vertex representation from a polyhedron (half-edge data structure)
     */
	public SharedVertexRepresentation(Polyhedron_3<Point_3> polyhedron){
		throw new Error("Warning: not supported in the minimal implementation of the Jcg library");
    	/*System.out.print("Creating a shared vertex representation from polyhedron...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

    	this.sizeVertices=polyhedron.sizeOfVertices();    
    	this.sizeHalfedges=polyhedron.sizeOfHalfedges();
    	
    	this.points=new Point_3[sizeVertices];
    	this.faceDegrees=new int[sizeFaces];
    	this.faces=new int[sizeFaces][];
    	
    	// store vertex indices in a hash table (for efficient retrieval)
    	HashMap<Vertex<Point_3>,Integer> vertexIndices=new HashMap<Vertex<Point_3>, Integer>();
    	
    	int i=0;
    	for(Vertex<Point_3> v: polyhedron.vertices) {
    		//points[i]=new Point_3(v.index);
    		vertexIndices.put(v, i);
    		i++;
    	}
      	
        endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");*/
  }
	
    /**
     * Store the representation in OFF format
     */
	public void writeOffFile(String filename) throws IOException {
    	// store vertex indices in map 
    	 BufferedWriter out = new BufferedWriter (new FileWriter(filename));
       	 out.write ("OFF\n");
       	 out.write (sizeVertices + " " + sizeFaces + " 0\n");  
    	 for (Point_3 p : points)
    		 out.write(p.getX() + " " + p.getY() + " " + p.getZ() + "\n");
    	 
    	 for (int i=0; i<sizeFaces; i++) {
    		 out.write(""+faceDegrees[i]);
    		 for (int j=0; j<faceDegrees[i]; j++)
    			 out.write (" " + faces[i][j]);
    			 out.write ("\n");
    	 }
    	 out.close();
	}
	
}
