package Jcg.polyhedron;

import java.util.*;

import Jcg.geometry.*;


/**
 * Half-Edge data structure for polyhedral meshes <p>
 * This is a 'minimal' space-efficient implementation: faces are not represented
 *
 * @author Code by Luca Castelli Aleardi (Ecole Polytechnique, INF562/INF555, 2010-2017)
 *
 * @see Primitives for the Manipulation of an orientable surface
 */
public class Polyhedron_3<X extends Point_>{

	public float[] coordF; // geometric coordinates (for testing performances without java interfaces)
    public ArrayList<Vertex<X>> vertices;
    public ArrayList<Halfedge<X>> halfedges;

    public Polyhedron_3() {
        vertices=new ArrayList<Vertex<X>>();
        halfedges=new ArrayList<Halfedge<X>>();
    }
   
    public Polyhedron_3(int n, int e, int f) {
        vertices=new ArrayList<Vertex<X>>(n);
        halfedges=new ArrayList<Halfedge<X>>(e);
		// geometric coordinates
		this.coordF=new float[n*3]; // coordinates simple float precision
    }
    
    class ColorDecorator extends Decorator<Vertex<X>,Integer> {
    }
    
    public void DecorateVertices() {
    	ColorDecorator color=new ColorDecorator();
    	color.setDecoration(vertices.get(0),0);
    }

    public int sizeOfVertices() {
    	return this.vertices.size();
    }
    
    public int sizeOfHalfedges() {
    	return this.halfedges.size();
    }
    
    /**
     * compute the degree of a vertex
     */
    public int vertexDegree(Vertex<X> v) {
    	int result=0;
    	Halfedge<X> e=v.getHalfedge();
    	
    	Halfedge<X> pEdge=e.getNext().getOpposite();
    	while(pEdge!=e) {
    		pEdge=pEdge.getNext().getOpposite();
    		result++;
    	}
    	
    	return result+1;
    }

/**
 * returns true if all vertices have exactly two incident edges
 */	
	public boolean isPureBivalent() {
		throw new Error("To be completed");
	}

/**
 * returns true if all vertices have exactly three incident edges
 */	
	public boolean isPureTrivalent() {
		throw new Error("a completer");
	}

	/**
	 * true iff the every face is a triangle
	 */	
	public boolean isPureTriangle() {
		for(Halfedge<X> h: this.halfedges) {
			if(h.next.next.next!=h)
				return false;
		}
		return true;
	}

/**
 * returns true if the polyhedral surface is combinatorially consistent. 
 * If borders==true normalization of the border edges is checked too. 
 * This method checks that each facet is at least a triangle and that 
 * the two incident facets of a non-border edge are distinct. 
 */	    
    public boolean isValid(boolean borders) {
    	boolean valid=true;
        System.out.print("Checking Polyhedron...");
        int n=this.vertices.size();
        int e=this.halfedges.size();

        int i=0;
        for(Halfedge<X> pedge: this.halfedges) {
        	if(pedge.getOpposite()==null || pedge.opposite.opposite!=pedge) { 
        		System.out.print("error opposite: "+i); 
        		valid=false; 
        	}
        	if(pedge.getNext()==null || pedge.next.getPrev()!=pedge) { 
        		System.out.println("error next_edge: "+i); 
        		valid=false; 
        	}
        	if(pedge.getPrev()==null || pedge.getPrev().next!=pedge) { 
        		System.out.println("error prev_edge: "+i); 
        		valid=false; 
        	}
        	if(pedge.getVertex()==null) { 
        		System.out.println("error vertex: "+i); 
        		valid=false;
        	}
        	if(pedge.opposite==null || pedge.opposite.opposite==null || pedge.opposite.opposite!=pedge) {
        		System.out.println("error opposite edge: "+i); 
        		valid=false;        		
        	}
        	i++;
        }            
        
        i=0;
        for(Vertex<X> pvertex: this.vertices){
        	if(pvertex==null) { System.out.println("error vertex pointer:"+i); valid=false; }
        	if(pvertex.halfedge==null) { System.out.println("error vertex.halfedge: "+i); valid=false; }
        	if(this.coordF==null) { System.out.println("error vertex.point: "+i); valid=false; }
        	if(pvertex.halfedge.vertex!=pvertex) { System.out.println("error vertex.halfedge: "+i); valid=false; }
        	i++;
        }            

        if(valid==true)
        	System.out.println("ok"); 
        else
         	System.out.println("not valid"); 
        
        if(this.isPureTriangle())
        	System.out.println("The mesh is pure triangle");
        else
        	System.out.println("The mesh is polygonal");
        
        System.out.println("n: "+n+"  e: "+e/2);
        
        return valid;
    }
    
    /**
     * Reset indices for all: vertices, faces and halfedges. <p>
     * Indices correspond to the order of storage in the collections (e.g., vertex v0 is the first in the ArrayList 'vertices')
     */	
    public void resetMeshIndices() {
    	int count=0;
    	for(Vertex v: this.vertices) {
    		if(v==null)
    			throw new Error("Error: null vertex reference in array 'vertices'");
    		v.index=count;
    		count++;
    	}

    	count=0;
    	for(Halfedge e: this.halfedges) {
       		if(e==null)
    			throw new Error("Error: null halfedge reference in array 'halfedges'");
    		e.index=count;
    		count++;
    	}
    }

    /**
     * Return a string representing the list of vertices
     */
    public String verticesToString() {
        String result="List of vertices\n";
        Iterator it=this.vertices.iterator();
        int cont=0;
        while(it.hasNext()) {
        	Vertex<X> v=(Vertex<X>)it.next();
        	result=result+this.vertexToString(v)+"\n"; 
        	cont++;
        }  	
        return result;           
    }

    /**
     * Return a string representing the list of vertices
     */
    public String vertexToString(Vertex v) {
        String result="v"+v.index+" ("+this.getX(v.index)+", "+this.getY(v.index)+", "+this.getZ(v.index)+")";
        return result;           
    }

    /**
     * Return a string representing the list of faces
     */
    public String edgesToString() {
        String result="List of edges\n";
        Iterator it=this.halfedges.iterator();
        int cont=0;
        while(it.hasNext()) {
        	Halfedge<X> e=(Halfedge<X>)it.next();
        	result=result+"e"+cont+" ";
        	//result=result+vertices.indexOf(e.getVertex())+"\n";
        	result=result+e.toString()+"\n";
        	cont++;
        }  	
        return result;           
    }
    
	/** 
	 * Return the geometric coordinates of vertex v
	 * 
	 * @param v  the index of the vertex
	 **/
	public float[] getPoint(int v) {
		int index=3*v;
		return new float[] {this.coordF[index], this.coordF[index+1], this.coordF[index+2]};
	}

	/** 
	 * Set the geometric coordinates of vertex v
	 * 
	 * @param v  the index of the vertex
	 **/
	public void setPoint(int v, float x, float y, float z) {
		int index=3*v;
		this.coordF[index]=x;
		this.coordF[index+1]=y;
		this.coordF[index+2]=z;
	}

	/** Return the x-coordinate of vertex v **/
	public float getX(int v) { return this.coordF[v*3]; }

	/** Return the y-coordinate of vertex v **/
	public float getY(int v) { return this.coordF[v*3+1]; }

	/** Return the z-coordinate of vertex v **/
	public float getZ(int v) { return this.coordF[v*3+2]; }
  
}