package Jcg.schnyderwoods;

import Jcg.polyhedron.*;
import Jcg.geometry.Point_;
import Jcg.util.*;

/**
 * @author Luca Castelli Aleardi (2017, Ecole Polytechnique)
 * 
 * This class allows to compute a Schnyder wood of a planar triangulation.
 * It implements a "cut border": a simple cycle of edges defining a
 * topological disk, and provides basic operations such as vertex removal
 * (the vertex to be removed must belong to the cut border)
 * It works for planar triangle meshes without boundaries.
 * Faces are assumed to be ccw oriented (only for combinatorics)
 */
public class PlanarTriSchnyderWood extends EdgeOrientation {
    // definition of the root face and root vertices v0, v1, v2
	
    /** half-edge (v_0, v_1) oriented toward v_1, assuming the ccw orientation of faces */
    public Halfedge<Point_> rootEdge;
    public Vertex<Point_> v0, v1, v2; // vertices on the root face (the outer face, in the plane)

    // auxiliary information needed for computing a Schnyder wood
    private DLinkedList<Halfedge<Point_>> outerCycle; // edges defining the outer cycle of the planar map   
    private boolean[] isChord; // for half-edges, says whether is a chord
    private boolean[] isOnCutBorder; // say whether a vertex is lying on the cut-border
    
    /**
     * Construct the cut-border starting from the root edge (v0, v1)
     * At the beginning the cut-border contains edges (v2, v0) and (v1, v2)
     * Edges are ccw oriented around faces
     */
    public PlanarTriSchnyderWood(Polyhedron_3<Point_> polyhedron, Halfedge<Point_> rootEdge) {
    	System.out.print("Initialize Schnyder wood computation: ");
    	if(polyhedron==null)
    		throw new Error("error: null polyhedron");
    	if(rootEdge==null)
    		throw new Error("error: root edge null");
    	
    	int i=0;
    	for(i=0;i<polyhedron.sizeOfVertices();i++) {
    		Vertex<Point_> v=(Vertex<Point_>)polyhedron.vertices.get(i);
    		v.index=i;
    	}
    	
    	i=0;
    	for(Object o: polyhedron.halfedges) {
    		Halfedge<Point_> e=(Halfedge<Point_>)o;
    		e.index=i;
    		i++;
    	}
    	
    	this.outerCycle=new DLinkedList<Halfedge<Point_>>();
    	this.polyhedron=polyhedron;
    	this.rootEdge=rootEdge;    	
        this.v0=this.rootEdge.getOpposite().getVertex();
        this.v1=this.rootEdge.getVertex();
    	
    	this.edgeColor=new byte[this.polyhedron.sizeOfHalfedges()];
    	for(i=0;i<edgeColor.length;i++)
    		this.edgeColor[i]=-1;
    	
    	this.isChord=new boolean[this.polyhedron.sizeOfHalfedges()];
    	this.isWellOriented=new boolean[this.polyhedron.sizeOfHalfedges()];
    	this.isOnCutBorder=new boolean[this.polyhedron.sizeOfVertices()];
    	
    	Halfedge<Point_> edge10=this.rootEdge.getOpposite();
    	Halfedge<Point_> edge02=edge10.getNext();
    	Halfedge<Point_> edge21=edge02.getNext();
    	
    	this.v2=edge02.getVertex();
    	
    	// setting information concerning the root-edge
    	this.outerCycle.add(edge02.getOpposite());	
    	this.outerCycle.add(edge21.getOpposite());
    	this.isOnCutBorder[v1.index]=true;
    	this.isOnCutBorder[v0.index]=true; // LCA newly added
    	this.isOnCutBorder[v2.index]=true; // LCA newly added
    	
    	// set the orientation of the root edge
    	this.isWellOriented[rootEdge.index]=false;
    	this.isWellOriented[rootEdge.getOpposite().index]=true;
    	// set the color of the root edge
    	this.edgeColor[rootEdge.index]=0;
    	this.edgeColor[rootEdge.getOpposite().index]=0;
    	System.out.print("\n\t root face (v"+v0.index+", v"+v1.index+", v"+v2.index+")");
    	System.out.println("\n\t root edge (v"+this.rootEdge.getOpposite().getVertex().index+", v"+this.rootEdge.getVertex().index+")");
   }
    
    /**
     * It removes a vertex from the cut-border (vertex conquest)
     * It updates the cut-border, assigning color and orientation to edges
     * 
     * @param node The node in the list storing the given half-edge (whose vertex will be removed)
     * @return result return the list node containing the preceding halfedge (on the boundary cycle)
     */
    public DListNode<Halfedge<Point_>> vertexRemoval(DListNode<Halfedge<Point_>> node) {
    	if(node==null || this.outerCycle.isEmpty()) {
    		System.out.println("no more vertex to remove ");
    		return null;
    	}
    	
    	Halfedge<Point_> rightEdge=node.getElement(); // the vertex v of rightEdge will be removed
    	if(rightEdge==null) {
    		throw new Error("null reference: rightEdge");
    	}
    	if(rightEdge.getVertex()==this.v0) // vertex v0 cannot be removed
    		return node.getNext();
    	
    	Halfedge<Point_> leftEdge=node.getPrev().getElement();
    	if(leftEdge==null) {
    		throw new Error("null reference: leftEdge");
    	}
    	
    	// the vertex is incident to a triangle
    	if(rightEdge.getNext()==leftEdge) {
    		//count++;
    		return triangleRemoval(node);
    	}
    	// the vertex is incident to exactly 2 triangles
    	//if(rightEdge.getNext().getOpposite().getNext()==leftEdge)
    		//count++;

    	// if the vertex has still incident chords it cannot be removed
    	// then return next vertex on cut-border
    	/*if(this.hasIncidentChords[rightEdge.getVertex().index]>0)
    		return i+1; */
    	if(this.hasIncidentChords(node)==true)
    		return node.getNext();
    	
    	// general case: the vertex is incident to more than one triangle
    	
    	// process left and right edges incident to v on the cut-border
    	this.setOutgoingEdge1(rightEdge);
    	this.setOutgoingEdge0(leftEdge);
    	//this.boundary.remove(i);
    	
    	// v is not anymore on the cut-border
    	this.isOnCutBorder[rightEdge.getVertex().index]=false;
    	
    	// add new edges to the cut-border
    	//this.addToCutBorder(rightEdge.getPrev().getOpposite(), i-1);
    	this.setToCutBorder(rightEdge.getPrev().getOpposite());
    	node.setElement(rightEdge.getPrev().getOpposite());
    	DListNode<Halfedge<Point_>> previousNode=node.getPrev(); // the boundary edge preceding
   	
    	Halfedge<Point_> pEdge=rightEdge.getNext().getOpposite();
    	while(pEdge!=leftEdge.getOpposite()) {
    		//this.addToCutBorder(pEdge.getPrev().getOpposite(), i-1); // old version
    		this.setToCutBorder(pEdge.getPrev().getOpposite());
    		this.outerCycle.insertAfter(previousNode, pEdge.getPrev().getOpposite());
    		this.setIngoingEdge2(pEdge);
    		pEdge=pEdge.getNext().getOpposite();
    	}
    	DListNode<Halfedge<Point_>> result=previousNode.getNext();
    	this.outerCycle.delete(previousNode);
    	
    	return result;
    }

    /**
     * It removes a vertex incident to one triangle from the boundary cycle
     * @param node the list node storing the halfedge whose vertex will be removed
     */
    private DListNode<Halfedge<Point_>> triangleRemoval(DListNode<Halfedge<Point_>> node) {
    	Halfedge<Point_> rightEdge=node.getElement();
    	Halfedge<Point_> leftEdge=node.getPrev().getElement();

    	this.setOutgoingEdge1(rightEdge);
    	this.setOutgoingEdge0(leftEdge);
    	//this.boundary.remove(i);
    	Halfedge<Point_> e=rightEdge.getPrev().getOpposite();
    	node.setElement(e);
    	this.outerCycle.delete(node.getPrev());

    	this.isOnCutBorder[rightEdge.getVertex().index]=false;
    	//this.addToCutBorder(rightEdge.getPrev().getOpposite(), i-1);
    	//this.boundary.add(position, e);

    	this.edgeColor[e.index]=3;
    	this.edgeColor[e.getOpposite().index]=3;
    	
    	this.isChord[e.index]=false;
    	this.isChord[e.getOpposite().index]=false;
    	
    	this.isOnCutBorder[e.getVertex().index]=true;
    	this.isOnCutBorder[e.getOpposite().getVertex().index]=true;

    	return node;
    }
    
    /**
     * Check whether a vertex has incident chords
     * In that case it cannot be removed from the cut-border
     */
    private boolean hasIncidentChords(DListNode<Halfedge<Point_>> node) {
    	Halfedge<Point_> rightEdge=node.getElement();
    	Halfedge<Point_> leftEdge=node.getPrev().getElement();

    	Halfedge<Point_> pEdge=rightEdge.getNext().getOpposite();
    	while(pEdge!=leftEdge.getOpposite()) {
    		Vertex<Point_> v=pEdge.getOpposite().getVertex();
    		if(this.isOnCutBorder[v.index]==true)
    			return true;
    		pEdge=pEdge.getNext().getOpposite();
    	}
    	return false;
    }

    private void setIngoingEdge2(Halfedge<Point_> e) {
    	this.edgeColor[e.index]=2;
    	this.edgeColor[e.getOpposite().index]=2;
    	
    	this.isWellOriented[e.index]=true;
    	this.isWellOriented[e.getOpposite().index]=false;
    }

    private void setOutgoingEdge1(Halfedge<Point_> e) {
    	this.edgeColor[e.index]=1;
    	this.edgeColor[e.getOpposite().index]=1;
    	
    	this.isWellOriented[e.index]=false;
    	this.isWellOriented[e.getOpposite().index]=true;
    }

    private void setOutgoingEdge0(Halfedge<Point_> e) {
    	this.edgeColor[e.index]=0;
    	this.edgeColor[e.getOpposite().index]=0;
    	
    	this.isWellOriented[e.index]=true;
    	this.isWellOriented[e.getOpposite().index]=false;
    }

    /**
     * Add a half-edge to the cut-border, at a given position (in the cut-border)
     * Update all concerned information (colors, existent chords, boundary vertices, ...)
     */
    public void addToCutBorder(Halfedge<Point_> e, DListNode<Halfedge<Point_>> node) {
    	//System.out.println("edge added to cut-border");
    	if(e==null)
    		throw new Error("halfedge not defined");
    	if(node==null)
    		throw new Error("error null reference: list node");
    	
    	this.outerCycle.insertBefore(node, e);
    	this.setToCutBorder(e);
    }

    /**
     * Add a half-edge to the cut-border, at a given position (in the cut-border)
     * Update all concerned information (colors, existent chords, boundary vertices, ...)
     */
    public void setToCutBorder(Halfedge<Point_> e) {
    	this.edgeColor[e.index]=3;
    	this.edgeColor[e.getOpposite().index]=3;
    	
    	this.isChord[e.index]=false;
    	this.isChord[e.getOpposite().index]=false;
    	
    	this.isOnCutBorder[e.getVertex().index]=true;
    	this.isOnCutBorder[e.getOpposite().getVertex().index]=true;
    }

    /**
     * Perform all steps of the vertex conquest, computing the Schnyder wood
     */
    public void performTraversal() {
    	System.out.print("Computing Schnyder wood (for planar triangulations)...");
    	//System.out.println("Cut border: \n"+this.toString());
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	DListNode<Halfedge<Point_>> node=this.outerCycle.getFirst().getNext();
    	if(node==null) {
    		throw new Error("error null reference: first boundary node not defined");
    	}
    	while(this.outerCycle.size()>1) {
    		//System.out.println("\nCut border: \n"+this.toString());
    		//System.out.println("Processing vertex "+node.getElement().getVertex().index);
    		node=this.vertexRemoval(node);
    	}
    	// set the color of the root edge, which is of color 0 (oriented toward v0)
    	this.edgeColor[rootEdge.index]=0;
    	this.edgeColor[rootEdge.getOpposite().index]=0;
        
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("done");
    	System.out.println(" ("+duration+" seconds)");
    	//System.out.println("Triangle removals: "+count);
    }
        
    /**
     * Return an array of indices representing the original vertex ordering
     */
    public int[] getOriginalVertexOrdering() {
    	int[] result=new int[this.polyhedron.sizeOfVertices()];
    	
    	int i=0;
    	for(Object o: this.polyhedron.vertices) {
    		Vertex<Point_> v=(Vertex<Point_>)o;
    		result[i]=v.index;
    		i++;
    	}
    	return result;    	
    }

	/**
	 * Return the first (half)edge oriented toward vertex v and having color 0 (red)
	 * Edges (v0, v1) and (v2, v0) are assumed to be red colored
	 * Return null, if such an edge does not exist
	 */
	public Halfedge<Point_> getFirstIncomingRedEdge(Vertex<Point_> v) {
		if(v==this.v0) {
			return this.rootEdge.getOpposite(); // half-edge (v1, v0), oriented toward v0
		}
		if(v==this.v1) // no incoming red edges
			return null;
		if(v==this.v2) // no incoming red edges
			return null;
		
    	// visit all (half)-edges incident to v, and the return the first incoming edge of color 0
		Halfedge<Point_> e=v.getHalfedge();
		if(this.edgeColor[e.index]==0 && this.isWellOriented[e.index]==true && this.edgeColor[e.getNext().index]==1)
			return e;
    	Halfedge<Point_> pEdge=e.getOpposite().getPrev();
    	while(pEdge!=e) { // turn around vertex v
    		if(this.edgeColor[pEdge.index]==0 && this.isWellOriented[pEdge.index]==true && this.edgeColor[pEdge.getNext().index]==1)
    			return pEdge;
    		pEdge=pEdge.getOpposite().getPrev();
    	}
    	return null;
	}

	/**
	 * Return the next edge of color 0, after edge e (turning in ccw direction), oriented toward its target vertex (denoted by v).
	 * Half-edge e is assumed to have v as target vertex
	 * Edges (v0, v1) and (v2, v0) are assumed to be red colored (and oriented toward v0)
	 * Return null, if such an edge does not exist
	 */
	public Halfedge<Point_> getNextIncomingRedEdge(Halfedge<Point_> e) {
		if(e==null) return null; 
		Halfedge<Point_> e20=this.rootEdge.getOpposite().getPrev(); // halfedge e=(v2, v0), having v0 as target
		if(e==e20) {
			return null; // e20 is the last ingoing edge (of color 0) incident to v0
		}
		
    	Halfedge<Point_> pEdge=e.getOpposite().getPrev();
    	if(this.edgeColor[pEdge.index]==0 && this.isWellOriented[pEdge.index]==true)
    			return pEdge;
    	return null;
	}
	
	/**
	 * Return the parenthesis encoding of the Schnyder wood, consisting of two words
	 * balanced parenthesis words.
	 * 
	 * It performs the traversal of the contour of the tree, in ccw order.
	 * The traversal starts from the half-edge (u, v0) and ends at (v0, v2).
	 * (u, v0) is the half-edge oriented toward v0 corresponding to
	 * the first incoming edge incident to v0 (thus of color 0)
	 */
	public String[] encodeSchnyderWood() {
		System.out.print("Encoding the triangulation (endowed with a Schnyder wood)...");
		String treeCode=""; // add first edge (v1, v0)
		String incomingDegrees=""; // vertex v1 has no incident incoming edges of color 2
		Halfedge<Point_> firstEdge=this.rootEdge.getPrev(); // starting half-edge, of color 0, oriented toward v_0
		Halfedge<Point_> edge20=this.rootEdge.getOpposite().getNext().getOpposite(); // halfedge v20, oriented toward v_0
		Halfedge<Point_> lastEdge=edge20.getOpposite();
		Halfedge<Point_> pEdge=firstEdge; 
		
    	while(pEdge!=lastEdge) { // perform the traversal of the contour of T_0
    		if(this.edgeColor[pEdge.index]==0 && this.isWellOriented[pEdge.index]==true) { // an edge of T_0 visited the first time
    			if(pEdge!=edge20)
    				treeCode=treeCode+"("; // do not write '(' for the edge (v2, v0)
    			if(pEdge!=firstEdge)
    				incomingDegrees=incomingDegrees+'['; // do not write '[' for the first edge oriented toward v0
    			pEdge=pEdge.getPrev();
    		}
    		else if(this.edgeColor[pEdge.index]==2 && this.isWellOriented[pEdge.index]==true) { // a black incoming edge
    			incomingDegrees=incomingDegrees+']';
    			pEdge=pEdge.getOpposite().getPrev();
    		}
    		else if(this.edgeColor[pEdge.index]==1 && this.isWellOriented[pEdge.index]==false) { // outoing blue edge
    			pEdge=pEdge.getOpposite().getPrev();
    		}
    		else if(this.edgeColor[pEdge.index]==2 && this.isWellOriented[pEdge.index]==false) { // black outoing edge
    			pEdge=pEdge.getOpposite().getPrev();
    		}
    		else if(this.edgeColor[pEdge.index]==1 && this.isWellOriented[pEdge.index]==true) { // a blue incoming edge
    			pEdge=pEdge.getOpposite().getPrev();
    		}
    		else if(this.edgeColor[pEdge.index]==0 && this.isWellOriented[pEdge.index]==false) { // an edge of T_0 encountered the second time
    			treeCode=treeCode+")";
    			pEdge=pEdge.getPrev();
    		}
    		else
    			throw new Error("Error: wrong edge orientation/coloration");
    	}
    	treeCode=treeCode+""; // add last closing parenthesis, corresponding to edge (v2, v0)
    	
    	System.out.println("done");
    	String[] result=new String[2];
    	result[0]=treeCode;
    	result[1]=incomingDegrees;
    	
    	
    	return result;
	}

    //----------------------------------------------
    //--- Methods for visualizing the cut-border ---
    //----------------------------------------------
    
    public String[] originalVertexOrderingToString() {
    	String[] result=new String[this.polyhedron.sizeOfVertices()];
    	
    	int i=0;
    	for(Object o: this.polyhedron.vertices) {
    		Vertex<Point_> v=(Vertex<Point_>)o;
    		result[i]=""+v.index;
    		i++;
    	}
    	return result;    	
    }

    public String toString() {
    	String result="cut-border size: "+this.outerCycle.size();
    	result=result+"\nroot edge (v0,v1)=("+v0.index+","+v1.index+")"+"   ("+v0+","+v1+")";
    	result=result+"\nroot face (v0,v1,v2)=("+v0+","+v1+","+v2+")";
    	
    	result=result+"\n boundary edges: [";
		DListNode<Halfedge<Point_>> n=this.outerCycle.getFirst();
		while(n!=this.outerCycle.getLast()) {
			result=result+n.getElement().getVertex().index+" ";
			n=n.getNext();
		}
		result=result+n.getElement().getVertex().index+" ";
		return result+"]";
    }
    
	/**
	 * Return a string representing the Schnyder wood
	 */
    public String SchnyderWoodToString() {
    	String result="\nroot edge (v0,v1)=("+v0.index+","+v1.index+")"+"   ("+v0+","+v1+")";
    	result=result+"\nroot face (v0,v1,v2)=("+v0+","+v1+","+v2+")\n";
    	
    	return result+this.orientationToString();
    }
    
}
